""" One of the widgets that main window composes of.

    See :class:`.MainWindow`
    Similar modules: class:`.AddCluster`, :class:`.RemoveCluster`,
    :class:`.IdactNotebook`, :class:`.AdjustTimeouts`
    Helpers: class:`.ShowJobsWindow`
"""
from PyQt5.QtWidgets import QWidget, QTableWidgetItem

from idact.detail.slurm.run_scancel import run_scancel
from idact.detail.slurm.run_squeue import run_squeue
from idact import show_cluster, load_environment

from gui.functionality.popup_window import WindowType, PopUpWindow
from gui.helpers.ui_loader import UiLoader
from gui.helpers.worker import Worker
from gui.helpers.custom_exceptions import NoClustersError


class ManageJobs(QWidget):
    """ Module of GUI that is responsible for allowing the management of jobs
    in the slurm queue.
    """

    def __init__(self, data_provider, parent=None):
        super().__init__(parent=parent)
        self.ui = UiLoader.load_ui_from_file('manage-jobs.ui', self)

        self.parent = parent
        self.data_provider = data_provider
        self.cluster_names = self.data_provider.get_cluster_names()
        self.popup_window = PopUpWindow()
        self.cluster = None

        self.ui.show_jobs_button.clicked.connect(self.concurrent_show_jobs)
        self.ui.refresh_button.clicked.connect(self.concurrent_show_jobs)
        self.ui.cancel_job_button.clicked.connect(self.concurrent_cancel_job)

        self.data_provider.remove_cluster_signal.connect(self.handle_cluster_list_modification)
        self.data_provider.add_cluster_signal.connect(self.handle_cluster_list_modification)
        self.ui.cluster_names_box.addItems(self.cluster_names)

        self.ui.cancel_job_button.setEnabled(False)
        self.ui.jobs_table.itemSelectionChanged.connect(
            lambda: self.ui.cancel_job_button.setEnabled(
                len(self.ui.jobs_table.selectedIndexes()) > 0))

    def concurrent_show_jobs(self):
        """ Setups the worker that allows to run the show_jobs functionality
        in the parallel thread.
        """
        self.ui.show_jobs_button.setEnabled(False)
        self.ui.refresh_button.setEnabled(False)

        worker = Worker(self.show_jobs)
        worker.signals.result.connect(self.handle_complete_show_jobs)
        worker.signals.error.connect(self.handle_error_show_jobs)
        self.parent.threadpool.start(worker)

    def handle_complete_show_jobs(self, jobs):
        """ Handles the completion of show jobs function.
        """
        self.ui.show_jobs_button.setEnabled(True)
        self.ui.refresh_button.setEnabled(True)

        counter = len(jobs)
        self.ui.jobs_table.setRowCount(counter)

        for i in range(counter):
            self.ui.jobs_table.setItem(i, 0, QTableWidgetItem(str(jobs[i].job_id)))
            self.ui.jobs_table.setItem(i, 1, QTableWidgetItem(str(jobs[i].end_time)))
            self.ui.jobs_table.setItem(i, 2, QTableWidgetItem(str(jobs[i].node_count)))
            self.ui.jobs_table.setItem(
                i, 3, QTableWidgetItem(','.join(jobs[i].node_list)) if jobs[i].node_list else '')
            self.ui.jobs_table.setItem(
                i, 4, QTableWidgetItem(str(jobs[i].reason if jobs[i].reason else '')))
            self.ui.jobs_table.setItem(i, 5, QTableWidgetItem(jobs[i].state))

    def handle_error_show_jobs(self, exception):
        """ Handles the error thrown while showing jobs.

            :param exception: Instance of the exception.
        """
        self.ui.show_jobs_button.setEnabled(True)
        self.ui.refresh_button.setEnabled(True)

        if isinstance(exception, NoClustersError):
            self.popup_window.show_message("There are no added clusters", WindowType.error)
        elif isinstance(exception, KeyError):
            self.popup_window.show_message("The cluster does not exist", WindowType.error)
        else:
            self.popup_window.show_message("An error occurred while listing jobs", WindowType.error, exception)

    def show_jobs(self):
        """ Main function responsible for showing jobs.
        """
        load_environment()
        cluster_name = str(self.ui.cluster_names_box.currentText())
        if not cluster_name:
            raise NoClustersError()
        self.cluster = show_cluster(name=cluster_name)
        node = self.cluster.get_access_node()
        jobs = list(run_squeue(node).values())
        return jobs

    def concurrent_cancel_job(self):
        """ Setups the worker that allows to run the cancel_job functionality
        in the parallel thread.
        """
        self.ui.cancel_job_button.setEnabled(False)

        worker = Worker(self.cancel_job)
        worker.signals.result.connect(self.handle_complete_cancel_job)
        worker.signals.error.connect(self.handle_error_cancel_job)
        self.parent.threadpool.start(worker)

    def handle_complete_cancel_job(self):
        """ Handles the completion of cancel job function.
        """
        self.ui.cancel_job_button.setEnabled(True)
        self.popup_window.show_message("Cancel command has been successfully executed", WindowType.success)
        self.concurrent_show_jobs()

    def handle_error_cancel_job(self, exception):
        """ Handles the error thrown while cancelling job.

            :param exception: Instance of the exception.
        """
        self.ui.cancel_job_button.setEnabled(True)

        if isinstance(exception, KeyError):
            self.popup_window.show_message("The cluster does not exist", WindowType.error)
        else:
            self.popup_window.show_message("An error occurred while cancelling job", WindowType.error)

        self.ui.cancel_job_button.setEnabled(True)

    def cancel_job(self):
        """ Main function responsible for cancelling a job.
        """
        self.ui.cancel_job_button.setEnabled(False)

        load_environment()
        node = self.cluster.get_access_node()

        indexes = self.ui.jobs_table.selectedIndexes()
        for index in sorted(indexes, reverse=True):
            job_id = int(self.ui.jobs_table.item(index.row(), 0).text())
            run_scancel(job_id, node)

    def handle_cluster_list_modification(self):
        """ Handles the modification of the clusters list.
        """
        self.cluster_names = self.data_provider.get_cluster_names()
        self.ui.cluster_names_box.clear()
        self.ui.cluster_names_box.addItems(self.cluster_names)



""" Main module of the GUI application. It composes all the widgets
into one window.

    See: class:`.AddCluster`, :class:`.RemoveCluster`,
    :class:`.IdactNotebook`, :class:`.AdjustTimeouts`,
    :class:`.ManageJobs`
"""
import sys

from PyQt5.QtCore import QThreadPool
from PyQt5.QtWidgets import QMainWindow

from gui.functionality.add_cluster import AddCluster
from gui.functionality.adjust_timeouts import AdjustTimeouts
from gui.functionality.idact_notebook import IdactNotebook
from gui.functionality.manage_jobs import ManageJobs
from gui.functionality.program_info_window import ProgramInfoWindow
from gui.functionality.help_window import HelpWindow
from gui.functionality.remove_cluster import RemoveCluster
from gui.functionality.show_logs_window import ShowLogsWindow
from gui.helpers.data_provider import DataProvider
from gui.helpers.ui_loader import UiLoader


class MainWindow(QMainWindow):
    def __init__(self):
        self.show_logs_window = ShowLogsWindow()
        sys.stdout = self.show_logs_window
        sys.stderr = self.show_logs_window

        super(MainWindow, self).__init__()
        self.ui = UiLoader.load_ui_from_file('main-window.ui', self)

        self.data_provider = DataProvider()
        self.program_info_window = ProgramInfoWindow()
        self.help_window = HelpWindow()
        self.actions_file_name = None
        self.threadpool = QThreadPool()

        self.setCentralWidget(IdactNotebook(self.data_provider, self))

        self.ui.deploy_notebook_action.triggered.connect(self.handle_deploy_notebook_action)
        self.ui.manage_jobs_action.triggered.connect(self.handle_manage_jobs_action)

        self.ui.add_cluster_action.triggered.connect(self.handle_add_cluster_action)
        self.ui.remove_cluster_action.triggered.connect(self.handle_remove_cluster_action)
        self.ui.edit_configuration_action.triggered.connect(self.handle_edit_configuration_action)

        self.ui.show_logs_action.triggered.connect(self.handle_show_logs_action)

        self.ui.see_help_action.triggered.connect(self.handle_see_help_action)
        self.ui.about_the_program_action.triggered.connect(self.handle_about_the_program_action)

    def handle_deploy_notebook_action(self):
        self.setCentralWidget(IdactNotebook(self.data_provider, self))

    def handle_manage_jobs_action(self):
        self.setCentralWidget(ManageJobs(self.data_provider, self))

    def handle_add_cluster_action(self):
        self.setCentralWidget(AddCluster(self.data_provider, self))

    def handle_remove_cluster_action(self):
        self.setCentralWidget(RemoveCluster(self.data_provider, self))

    def handle_edit_configuration_action(self):
        self.setCentralWidget(AdjustTimeouts(self.data_provider, self))

    def handle_show_logs_action(self):
        self.show_logs_window.show()

    def handle_see_help_action(self):
        self.help_window.show()

    def handle_about_the_program_action(self):
        self.program_info_window.show()



""" One of the widgets that main window composes of.

    See :class:`.MainWindow`
    Similar modules: class:`.AddCluster`, :class:`.RemoveCluster`,
    :class:`.AdjustTimeouts`, :class:`.ManageJobs`
    Helpers: class:`.EditNativeArgumentsWindow`
"""
from PyQt5.QtWidgets import QWidget, QTableWidgetItem, QMessageBox
from contextlib import ExitStack

from idact import load_environment, show_cluster, Walltime
from idact.detail.config.client.client_cluster_config import ClusterConfigImpl
from idact.detail.deployment.cancel_local_on_exit import cancel_local_on_exit
from idact.detail.deployment.cancel_on_exit import cancel_on_exit
from idact.detail.jupyter_app.app_allocation_parameters import \
    AppAllocationParameters
from idact.detail.jupyter_app.override_parameters_if_possible import \
    override_parameters_if_possible
from idact.detail.jupyter_app.sleep_until_allocation_ends import \
    sleep_until_allocation_ends

from gui.functionality.popup_window import WindowType, PopUpWindow
from gui.functionality.yes_or_no_window import YesOrNoWindow
from gui.helpers.native_saver import NativeArgsSaver
from gui.helpers.parameter_saver import ParameterSaver
from gui.helpers.ui_loader import UiLoader
from gui.helpers.worker import Worker
from gui.helpers.custom_exceptions import NoClustersError


class IdactNotebook(QWidget):
    """ Module of GUI that is responsible for deploying the notebook
    on the selected cluster.
    """

    def __init__(self, data_provider, parent=None):
        super().__init__(parent=parent)
        self.ui = UiLoader.load_ui_from_file('deploy-notebook.ui', self)

        self.parent = parent
        self.data_provider = data_provider

        self.edit_native_arguments_window = EditNativeArgumentsWindow()
        self.popup_window = PopUpWindow()
        self.native_args_saver = NativeArgsSaver()
        self.saver = ParameterSaver()
        self.parameters = self.saver.get_map()

        self.ui.deploy_button.clicked.connect(self.concurrent_deploy_notebook)
        self.ui.edit_native_arguments_button.clicked.connect(self.open_edit_native_argument)

        self.ui.nodes_edit.setValue(self.parameters['deploy_notebook_arguments']['nodes'])
        self.ui.cores_edit.setValue(self.parameters['deploy_notebook_arguments']['cores'])
        self.ui.memory_edit.setValue(int(self.parameters['deploy_notebook_arguments']['memory_value']))
        self.ui.memory_unit_box.setCurrentText(self.parameters['deploy_notebook_arguments']['memory_unit'])
        self.ui.walltime_edit.setText(self.parameters['deploy_notebook_arguments']['walltime'])

        self.edit_native_arguments_window.ui.add_argument_button.clicked.connect(self.add_argument_row)
        self.edit_native_arguments_window.ui.remove_arguments_button.clicked.connect(self.remove_arguments)
        self.edit_native_arguments_window.ui.table_widget.itemSelectionChanged.connect(
            lambda: self.edit_native_arguments_window.ui.remove_arguments_button.setEnabled(
                len(self.edit_native_arguments_window.ui.table_widget.selectedIndexes()) > 0))
        self.edit_native_arguments_window.ui.save_arguments_button.clicked.connect(self.save_arguments)

        self.cluster_names = self.data_provider.get_cluster_names()

        self.data_provider.remove_cluster_signal.connect(self.handle_cluster_list_modification)
        self.data_provider.add_cluster_signal.connect(self.handle_cluster_list_modification)
        self.ui.cluster_names_box.addItems(self.cluster_names)

    def concurrent_deploy_notebook(self):
        """ Setups the worker that allows to run the deploy_notebook functionality
        in the parallel thread.
        """
        worker = Worker(self.deploy_notebook)
        worker.signals.result.connect(self.handle_complete_deploy_notebook)
        worker.signals.error.connect(self.handle_error_deploy_notebook)
        self.parent.threadpool.start(worker)

    def handle_complete_deploy_notebook(self):
        """ Handles the completion of deploy of the notebook.
        """
        self.popup_window.show_message("Notebook has been closed", WindowType.success)

    def handle_error_deploy_notebook(self, exception):
        """ Handles the error thrown while deploying the notebook.

            :param exception: Instance of the exception.
        """
        if isinstance(exception, NoClustersError):
            self.popup_window.show_message("There are no added clusters", WindowType.error)
        else:
            self.popup_window.show_message("An error occurred while deploying notebook", WindowType.error, exception)
        self.ui.deploy_button.setEnabled(True)

    def deploy_notebook(self):
        """ Main function responsible for deploying the notebook.
        """
        cluster_name = str(self.ui.cluster_names_box.currentText())

        if not cluster_name:
            raise NoClustersError()

        self.ui.deploy_button.setEnabled(False)

        nodes = int(self.ui.nodes_edit.text())
        self.parameters['deploy_notebook_arguments']['nodes'] = nodes
        cores = int(self.ui.cores_edit.text())
        self.parameters['deploy_notebook_arguments']['cores'] = cores
        memory = self.ui.memory_edit.text()
        self.parameters['deploy_notebook_arguments']['memory_value'] = memory
        unit = self.ui.memory_unit_box.currentText()
        self.parameters['deploy_notebook_arguments']['memory_unit'] = unit
        walltime = IdactNotebook.validate_and_format_walltime(self.ui.walltime_edit.text())
        self.parameters['deploy_notebook_arguments']['walltime'] = walltime
        self.saver.save(self.parameters)

        with ExitStack() as stack:
            load_environment()

            cluster = show_cluster(name=cluster_name)

            config = cluster.config
            assert isinstance(config, ClusterConfigImpl)
            parameters = AppAllocationParameters.deserialize(
                serialized=config.notebook_defaults)

            native_args = self.get_prepared_native_args()
            override_parameters_if_possible(parameters=parameters,
                                            nodes=nodes,
                                            cores=cores,
                                            memory_per_node=memory + unit,
                                            walltime=walltime,
                                            native_args=native_args.items())
            nodes = cluster.allocate_nodes(
                nodes=parameters.nodes,
                cores=parameters.cores,
                memory_per_node=parameters.memory_per_node,
                walltime=parameters.walltime,
                native_args=native_args)
            stack.enter_context(cancel_on_exit(nodes))
            nodes.wait()

            notebook = nodes[0].deploy_notebook()
            stack.enter_context(cancel_local_on_exit(notebook))

            cluster.push_deployment(nodes)

            cluster.push_deployment(notebook)

            notebook.open_in_browser()
            self.ui.deploy_button.setEnabled(True)
            sleep_until_allocation_ends(nodes=nodes)
        return

    @staticmethod
    def validate_and_format_walltime(walltime):
        """ Checks if walltime is correct and converts it to a proper format.

        :param walltime: Walltime string to check and convert.
                         Expected format: [days-]hours:minutes:seconds (for example 00:10:00)
                         or {days}d {hours}h {minutes}m {seconds}s - not all necessary and in any order
                         (for example 20m 30s)
        """
        if len(walltime) == 0:
            raise ValueError('Walltime cannot be empty.')

        if ':' not in walltime:
            walltime_elements = walltime.split(" ")

            days = 0
            hours = 0
            minutes = 0
            seconds = 0

            for element in walltime_elements:
                letter = element[len(element) - 1]
                number = element[0:len(element) - 1]
                if not number.isdigit():
                    raise ValueError('Wrong walltime format.')

                if letter == 'd':
                    if days != 0:
                        raise ValueError('Wrong walltime format. Days parameter is multiple defined.')
                    days = int(number)
                elif letter == 'h':
                    if hours != 0:
                        raise ValueError('Wrong walltime format. Hours parameter is multiple defined.')
                    hours = int(number)
                elif letter == 'm':
                    if minutes != 0:
                        raise ValueError('Wrong walltime format. Minutes parameter is multiple defined.')
                    minutes = int(number)
                elif letter == 's':
                    if seconds != 0:
                        raise ValueError('Wrong walltime format. Seconds parameter is multiple defined.')
                    seconds = int(number)
                else:
                    raise ValueError('Wrong walltime format.')

            walltime = str(Walltime(days=days, hours=hours, minutes=minutes, seconds=seconds))

        else:
            try:
                Walltime.from_string(walltime)
            except ValueError:
                raise ValueError('Wrong walltime format.')

        return walltime

    def get_prepared_native_args(self):
        """ Fetches the saved native_args from the native_args.json file.
        Converts them into proper format.
        """
        args = self.native_args_saver.get_native_args().copy()
        keys = list(args)

        for key in keys:
            if not key.startswith('--'):
                args['--' + key] = args.pop(key)
        return args

    def open_edit_native_argument(self):
        """ Opens the window with native arguments edition.
        """
        native_args = self.native_args_saver.get_native_args()
        counter = len(native_args)

        self.edit_native_arguments_window.ui.table_widget.setRowCount(counter)
        self.edit_native_arguments_window.ui.remove_arguments_button.setEnabled(False)

        row = 0
        for key in native_args.keys():
            self.edit_native_arguments_window.ui.table_widget.setItem(row, 0, QTableWidgetItem(key))
            self.edit_native_arguments_window.ui.table_widget.setItem(row, 1, QTableWidgetItem(native_args[key]))
            row += 1

        self.edit_native_arguments_window.show()
        self.edit_native_arguments_window.data_changed = False
        self.edit_native_arguments_window.ui.table_widget.cellChanged.connect(
            self.edit_native_arguments_window.set_that_data_changed)

    def add_argument_row(self):
        """ Adds the empty row to the native arguments table.
        """
        self.edit_native_arguments_window.ui.table_widget.setRowCount(
            self.edit_native_arguments_window.ui.table_widget.rowCount() + 1)

    def remove_arguments(self):
        """ Removes the selected native argument.
        """
        indexes = self.edit_native_arguments_window.ui.table_widget.selectedIndexes()

        for index in sorted(indexes, reverse=True):
            self.edit_native_arguments_window.ui.table_widget.removeRow(index.row())

        self.edit_native_arguments_window.set_that_data_changed()

    def save_arguments(self):
        """ Saves edited native arguments.
        """
        native_args = dict()

        for i in range(self.edit_native_arguments_window.ui.table_widget.rowCount()):
            name_item = self.edit_native_arguments_window.ui.table_widget.item(i, 0)
            value_item = self.edit_native_arguments_window.ui.table_widget.item(i, 1)

            name = None
            value = None
            if name_item is not None and value_item is not None:
                name = str(name_item.text())
                value = str(value_item.text())

            if not name or not value:
                self.popup_window.show_message("Name or value is empty", WindowType.error)
                return

            native_args[name] = value

        self.native_args_saver.save(native_args)
        self.edit_native_arguments_window.data_changed = False
        self.edit_native_arguments_window.close()

    def handle_cluster_list_modification(self):
        """ Handles the modification of the clusters list.
        """
        self.cluster_names = self.data_provider.get_cluster_names()
        self.ui.cluster_names_box.clear()
        self.ui.cluster_names_box.addItems(self.cluster_names)


class EditNativeArgumentsWindow(QWidget):
    """ Helper widget window for idact notebook native arguments.
    """

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.ui = UiLoader.load_ui_from_file('edit-native.ui', self)

        self.data_changed = False
        self.yes_or_no_window = YesOrNoWindow()

    def set_that_data_changed(self):
        self.data_changed = True

    def show_warning_window(self, event):
        """ Shows warning window that some changes may not have been saved.

            :param event: close event
        """
        close_window = self.yes_or_no_window.show_message(
            "Some changes may not have been saved. \nDo you want to quit anyway?")
        self.yes_or_no_window.box.setDefaultButton(QMessageBox.No)
        if not close_window:
            event.ignore()

    def closeEvent(self, QCloseEvent):
        """ Overriden method on closing Edit native arguments window.

            :param QCloseEvent: close event
        """
        if self.data_changed:
            self.show_warning_window(QCloseEvent)



""" One of the widgets that main window composes of.

    See :class:`.MainWindow`
    Similar modules:  class:`.AddCluster`, :class:`.RemoveCluster`,
    :class:`.IdactNotebook`, :class:`.ManageJobs`
"""
from PyQt5.QtWidgets import QWidget

from idact.core.retry import Retry
from idact.detail.environment.environment_provider import EnvironmentProvider
from idact import save_environment, load_environment

from gui.functionality.popup_window import WindowType, PopUpWindow
from gui.helpers.ui_loader import UiLoader


class AdjustTimeouts(QWidget):
    """ Module of GUI that is responsible for allowing the adjustment of the
    clusters timeouts.
    """

    def __init__(self, data_provider, parent=None):
        super().__init__(parent=parent)
        self.ui = UiLoader.load_ui_from_file('adjust-timeouts.ui', self)

        self.data_provider = data_provider
        self.popup_window = PopUpWindow()
        self.cluster_names = self.data_provider.get_cluster_names()
        self.current_cluster = ''

        load_environment()

        self.ui.cluster_names_box.addItems(self.cluster_names)

        if len(self.cluster_names) > 0:
            self.current_cluster = self.cluster_names[0]
            self.refresh_timeouts(self.current_cluster)
        else:
            self.current_cluster = ''

        self.ui.cluster_names_box.activated[str].connect(self.item_pressed)
        self.ui.save_timeouts_button.clicked.connect(self.save_timeouts)

        self.data_provider.add_cluster_signal.connect(self.handle_cluster_list_modification)
        self.data_provider.remove_cluster_signal.connect(self.handle_cluster_list_modification)

    def refresh_timeouts(self, cluster_name):
        """ Fetches and refreshes the timeouts of the particular cluster.

            :param cluster_name: Name of a cluster to refresh timeouts for.
        """
        load_environment()
        default_retries = EnvironmentProvider().environment.clusters[cluster_name].config.retries

        self.ui.port_info_count.setValue(default_retries[Retry.PORT_INFO].count)
        self.ui.jupyter_json_count.setValue(default_retries[Retry.JUPYTER_JSON].count)
        self.ui.scheduler_connect_count.setValue(default_retries[Retry.SCHEDULER_CONNECT].count)
        self.ui.dask_node_connect_count.setValue(default_retries[Retry.DASK_NODE_CONNECT].count)
        self.ui.deploy_dask_scheduler_count.setValue(default_retries[Retry.DEPLOY_DASK_SCHEDULER].count)
        self.ui.deploy_dask_worker_count.setValue(default_retries[Retry.DEPLOY_DASK_WORKER].count)
        self.ui.get_scheduler_address_count.setValue(default_retries[Retry.GET_SCHEDULER_ADDRESS].count)
        self.ui.check_worker_started_count.setValue(default_retries[Retry.CHECK_WORKER_STARTED].count)
        self.ui.cancel_deployment_count.setValue(default_retries[Retry.CANCEL_DEPLOYMENT].count)
        self.ui.squeue_after_sbatch_count.setValue(default_retries[Retry.SQUEUE_AFTER_SBATCH].count)
        self.ui.open_tunnel_count.setValue(default_retries[Retry.OPEN_TUNNEL].count)
        self.ui.validate_http_tunnel_count.setValue(default_retries[Retry.VALIDATE_HTTP_TUNNEL].count)
        self.ui.tunnel_try_again_with_any_port_count.setValue(
            default_retries[Retry.TUNNEL_TRY_AGAIN_WITH_ANY_PORT].count)

        self.ui.port_info_seconds.setValue(default_retries[Retry.PORT_INFO].seconds_between)
        self.ui.jupyter_json_seconds.setValue(default_retries[Retry.JUPYTER_JSON].seconds_between)
        self.ui.scheduler_connect_seconds.setValue(default_retries[Retry.SCHEDULER_CONNECT].seconds_between)
        self.ui.dask_node_connect_seconds.setValue(default_retries[Retry.DASK_NODE_CONNECT].seconds_between)
        self.ui.deploy_dask_scheduler_seconds.setValue(
            default_retries[Retry.DEPLOY_DASK_SCHEDULER].seconds_between)
        self.ui.deploy_dask_worker_seconds.setValue(default_retries[Retry.DEPLOY_DASK_WORKER].seconds_between)
        self.ui.get_scheduler_address_seconds.setValue(
            default_retries[Retry.GET_SCHEDULER_ADDRESS].seconds_between)
        self.ui.check_worker_started_seconds.setValue(default_retries[Retry.CHECK_WORKER_STARTED].seconds_between)
        self.ui.cancel_deployment_seconds.setValue(default_retries[Retry.CANCEL_DEPLOYMENT].seconds_between)
        self.ui.squeue_after_sbatch_seconds.setValue(default_retries[Retry.SQUEUE_AFTER_SBATCH].seconds_between)
        self.ui.open_tunnel_seconds.setValue(default_retries[Retry.OPEN_TUNNEL].seconds_between)
        self.ui.validate_http_tunnel_seconds.setValue(default_retries[Retry.VALIDATE_HTTP_TUNNEL].seconds_between)
        self.ui.tunnel_try_again_with_any_port_seconds.setValue(
            default_retries[Retry.TUNNEL_TRY_AGAIN_WITH_ANY_PORT].seconds_between)

    def save_timeouts(self):
        """ Saves to the configuration the timeouts of current cluster.
        """
        if self.current_cluster == '':
            self.popup_window.show_message("There are no added clusters", WindowType.error)
        else:
            default_retries = EnvironmentProvider().environment.clusters[self.current_cluster].config.retries

            default_retries[Retry.PORT_INFO].count = int(self.ui.port_info_count.text())
            default_retries[Retry.JUPYTER_JSON].count = int(self.ui.jupyter_json_count.text())
            default_retries[Retry.SCHEDULER_CONNECT].count = int(self.ui.scheduler_connect_count.text())
            default_retries[Retry.DASK_NODE_CONNECT].count = int(self.ui.dask_node_connect_count.text())
            default_retries[Retry.DEPLOY_DASK_SCHEDULER].count = int(self.ui.deploy_dask_scheduler_count.text())
            default_retries[Retry.DEPLOY_DASK_WORKER].count = int(self.ui.deploy_dask_worker_count.text())
            default_retries[Retry.GET_SCHEDULER_ADDRESS].count = int(self.ui.get_scheduler_address_count.text())
            default_retries[Retry.CHECK_WORKER_STARTED].count = int(self.ui.check_worker_started_count.text())
            default_retries[Retry.CANCEL_DEPLOYMENT].count = int(self.ui.cancel_deployment_count.text())
            default_retries[Retry.SQUEUE_AFTER_SBATCH].count = int(self.ui.squeue_after_sbatch_count.text())
            default_retries[Retry.OPEN_TUNNEL].count = int(self.ui.open_tunnel_count.text())
            default_retries[Retry.VALIDATE_HTTP_TUNNEL].count = int(self.ui.validate_http_tunnel_count.text())
            default_retries[Retry.TUNNEL_TRY_AGAIN_WITH_ANY_PORT].count = int(
                self.ui.tunnel_try_again_with_any_port_count.text())

            default_retries[Retry.PORT_INFO].seconds_between = int(self.ui.port_info_seconds.text())
            default_retries[Retry.JUPYTER_JSON].seconds_between = int(self.ui.jupyter_json_seconds.text())
            default_retries[Retry.SCHEDULER_CONNECT].seconds_between = int(self.ui.scheduler_connect_seconds.text())
            default_retries[Retry.DASK_NODE_CONNECT].seconds_between = int(self.ui.dask_node_connect_seconds.text())
            default_retries[Retry.DEPLOY_DASK_SCHEDULER].seconds_between = int(
                self.ui.deploy_dask_scheduler_seconds.text())
            default_retries[Retry.DEPLOY_DASK_WORKER].seconds_between = int(self.ui.deploy_dask_worker_seconds.text())
            default_retries[Retry.GET_SCHEDULER_ADDRESS].seconds_between = int(
                self.ui.get_scheduler_address_seconds.text())
            default_retries[Retry.CHECK_WORKER_STARTED].seconds_between = int(
                self.ui.check_worker_started_seconds.text())
            default_retries[Retry.CANCEL_DEPLOYMENT].seconds_between = int(self.ui.cancel_deployment_seconds.text())
            default_retries[Retry.SQUEUE_AFTER_SBATCH].seconds_between = int(self.ui.squeue_after_sbatch_seconds.text())
            default_retries[Retry.OPEN_TUNNEL].seconds_between = int(self.ui.open_tunnel_seconds.text())
            default_retries[Retry.VALIDATE_HTTP_TUNNEL].seconds_between = int(
                self.ui.validate_http_tunnel_seconds.text())
            default_retries[Retry.TUNNEL_TRY_AGAIN_WITH_ANY_PORT].seconds_between = int(
                self.ui.tunnel_try_again_with_any_port_seconds.text())

            save_environment()
            self.popup_window.show_message("Timeouts have been saved", WindowType.success)

    def item_pressed(self, item_pressed):
        """ Handles the cluster change.

            :param item_pressed: Name of a cluster that was selected.
        """
        self.current_cluster = item_pressed
        self.refresh_timeouts(item_pressed)

    def handle_cluster_list_modification(self):
        """ Handles the modification of the clusters list.
        """
        self.cluster_names = self.data_provider.get_cluster_names()
        self.ui.cluster_names_box.clear()
        self.ui.cluster_names_box.addItems(self.cluster_names)



""" One of the helpers for the gui application.

    Similar modules: class:`.DataProvider`, :class:`.ParameterSaver`,
    :class:`.NativeArgsSaver`, :class:`.Worker`, :class:`.ConfigurationProvider`
"""
import os

from PyQt5 import uic


class UiLoader:
    """ Loads the ui for the widget.
    """

    @staticmethod
    def load_ui_from_file(file_name, form):
        """ Loads the ui for the widget.

            :param file_name: Name of ui file to be loaded.
            :param form: Instance of the widget.
        """
        ui_path = os.path.dirname(os.path.abspath(__file__))
        ui_path = os.path.join(ui_path, os.pardir, 'widgets_templates', file_name)
        return uic.loadUi(ui_path, form)



""" Basic tests for idact-gui.
"""
import pytest
from pytestqt.qt_compat import qt_api
from gui.helpers.configuration_provider import ConfigurationProvider
from gui.functionality.main_window import MainWindow
from gui.helpers.data_provider import DataProvider
from gui.functionality.idact_notebook import IdactNotebook
from gui.functionality.manage_jobs import ManageJobs
from gui.functionality.add_cluster import AddCluster
from gui.functionality.remove_cluster import RemoveCluster
from gui.functionality.adjust_timeouts import AdjustTimeouts


@pytest.fixture()
def window():
    assert qt_api.QApplication.instance() is not None
    conf_provider = ConfigurationProvider()
    if not conf_provider.check_if_conf_file_exists():
        conf_provider.create_conf_file()
    window = MainWindow()
    return window


def test_basics(window, qtbot):
    """ Tests if idact-gui renders itself.
    """
    window.show()

    assert window.isVisible()
    assert window.windowTitle() == 'Idact GUI'


def test_deploy_notebook_window(window, qtbot):
    """ Tests if it is possible to open deploy notebook window.
    """
    window.show()
    window.deploy_notebook_action.trigger()

    data_provider = DataProvider()
    assert window.centralWidget().__class__ == IdactNotebook(data_provider).__class__


def test_manage_jobs_window(window, qtbot):
    """ Tests if it is possible to open manage jobs window.
    """
    window.show()
    window.manage_jobs_action.trigger()
    data_provider = DataProvider()

    assert window.centralWidget().__class__ == ManageJobs(data_provider).__class__


def test_add_cluster_window(window, qtbot):
    """ Tests if it is possible to open add cluster window.
    """
    window.show()
    window.add_cluster_action.trigger()
    data_provider = DataProvider()

    assert window.centralWidget().__class__ == AddCluster(data_provider).__class__


def test_remove_cluster_window(window, qtbot):
    """ Tests if it is possible to open remove cluster window.
    """
    window.show()
    window.remove_cluster_action.trigger()
    data_provider = DataProvider()

    assert window.centralWidget().__class__ == RemoveCluster(data_provider).__class__


def test_edit_configuration_window(window, qtbot):
    """ Tests if it is possible to open edit configuration window.
    """
    window.show()
    window.edit_configuration_action.trigger()
    data_provider = DataProvider()

    assert window.centralWidget().__class__ == AdjustTimeouts(data_provider).__class__


def test_logs_window(window, qtbot):
    """ Tests if it is possible to open logs window.
    """
    window.show()
    window.show_logs_action.trigger()

    assert window.show_logs_window.isVisible()
    assert window.show_logs_window.windowTitle() == 'Logs'


def test_help_window(window, qtbot):
    """ Tests if it is possible to open help window.
    """
    window.show()
    window.see_help_action.trigger()

    assert window.help_window.isVisible()
    assert window.help_window.windowTitle() == 'Help'


def test_about_window(window, qtbot):
    """ Tests if it is possible to open help window.
    """
    window.show()
    window.about_the_program_action.trigger()

    assert window.program_info_window.isVisible()
    assert window.program_info_window.windowTitle() == 'About'



""" One of the widgets that main window composes of.

    See :class:`.MainWindow`
    Similar modules: :class:`.RemoveCluster`, :class:`.IdactNotebook`,
    :class:`.AdjustTimeouts`, :class:`.ManageJobs`
"""
from PyQt5.QtWidgets import QWidget, QLineEdit, QFileDialog

from idact.core.auth import AuthMethod, KeyType
from idact.core.add_cluster import add_cluster
from idact.detail.config.client.setup_actions_config import SetupActionsConfigImpl
from idact.detail.add_cluster_app import actions_parser as parser
from idact import save_environment, load_environment

from gui.functionality.popup_window import WindowType, PopUpWindow
from gui.helpers.parameter_saver import ParameterSaver
from gui.helpers.ui_loader import UiLoader
from gui.helpers.worker import Worker
from gui.helpers.custom_exceptions import EmptyFieldError


class AddCluster(QWidget):
    """ Module of GUI that is responsible for allowing the addition of the
    new cluster.
    """

    def __init__(self, data_provider, parent=None):
        super().__init__(parent=parent)
        self.ui = UiLoader.load_ui_from_file('add-cluster.ui', self)

        self.parent = parent
        self.data_provider = data_provider

        self.popup_window = PopUpWindow()
        self.saver = ParameterSaver()
        self.parameters = self.saver.get_map()

        self.ui.password_edit.setEchoMode(QLineEdit.Password)
        self.ui.add_cluster_button.clicked.connect(self.concurrent_add_cluster)
        self.ui.cluster_name_addc_edit.setText(self.parameters['add_cluster_arguments']['cluster_name'])
        self.ui.user_edit.setText(self.parameters['add_cluster_arguments']['user'])
        self.ui.host_edit.setText(self.parameters['add_cluster_arguments']['host'])
        self.ui.port_edit.setValue(self.parameters['add_cluster_arguments']['port'])
        self.ui.add_actions_file_button.clicked.connect(self.open_actions_file_dialog)
        self.ui.clear_actions_file_button.clicked.connect(self.clear_actions_file)
        self.ui.add_key_button.clicked.connect(self.add_key)
        self.ui.clear_key_button.clicked.connect(self.clear_key)
        self.ui.auth_method_box.currentIndexChanged.connect(self.auth_method_toggled)
        self.actions_file_name = None
        self.key_path = None

    def auth_method_toggled(self):
        """ Handles the change of authentication method.
        """
        if self.ui.auth_method_box.currentText() == 'GENERATE_KEY':
            self.ui.password_label.setDisabled(False)
            self.ui.password_edit.setDisabled(False)
            self.ui.private_key_label.setDisabled(True)
            self.ui.currently_selected_key.setDisabled(True)
            self.ui.add_key_button.setDisabled(True)
            self.ui.clear_key_button.setDisabled(True)
        else:
            self.ui.password_label.setDisabled(True)
            self.ui.password_edit.setDisabled(True)
            self.ui.private_key_label.setDisabled(False)
            self.ui.currently_selected_key.setDisabled(False)
            self.ui.add_key_button.setDisabled(False)
            self.ui.clear_key_button.setDisabled(True)

        if self.key_path is not None:
            self.ui.clear_key_button.setDisabled(False)

    def concurrent_add_cluster(self):
        """ Setups the worker that allows to run the add_cluster functionality
        in the parallel thread.
        """
        self.ui.add_cluster_button.setDisabled(True)

        worker = Worker(self.add_cluster)
        worker.signals.result.connect(self.handle_complete_add_cluster)
        worker.signals.error.connect(self.handle_error_add_cluster)
        self.parent.threadpool.start(worker)

    def handle_complete_add_cluster(self):
        """ Handles the completion of adding cluster.
        """
        self.ui.add_cluster_button.setDisabled(False)
        save_environment()
        self.data_provider.add_cluster_signal.emit()
        self.popup_window.show_message("The cluster has been successfully added", WindowType.success)

    def handle_error_add_cluster(self, exception):
        """ Handles the error thrown while adding a cluster.

            :param exception: Instance of the exception.
        """
        self.ui.add_cluster_button.setDisabled(False)
        if isinstance(exception, EmptyFieldError):
            self.popup_window.show_message("Cluster name cannot be empty", WindowType.error)
        elif isinstance(exception, ValueError):
            self.popup_window.show_message("Cluster already exists", WindowType.error)
        else:
            self.popup_window.show_message("An error occurred while adding cluster", WindowType.error, exception)

    def add_cluster(self):
        """ Main function responsible for adding a cluster.
        """
        load_environment()

        cluster_name = self.ui.cluster_name_addc_edit.text()

        if not cluster_name:
            raise EmptyFieldError()

        self.parameters['add_cluster_arguments']['cluster_name'] = cluster_name
        user = self.ui.user_edit.text()
        self.parameters['add_cluster_arguments']['user'] = user
        host = self.ui.host_edit.text()
        self.parameters['add_cluster_arguments']['host'] = host
        port = int(self.ui.port_edit.text())
        self.parameters['add_cluster_arguments']['port'] = port
        self.saver.save(self.parameters)
        auth = self.ui.auth_method_box.currentText()
        password = self.ui.password_edit.text()
        use_jupyter_lab = self.ui.use_jupyter_lab_check_box.isChecked()

        setup_actions = None
        if self.actions_file_name is not None:
            setup_actions = SetupActionsConfigImpl()
            setup_actions.jupyter = parser.parse_actions(self.actions_file_name)

        if auth == 'GENERATE_KEY':
            cluster = add_cluster(name=cluster_name,
                                  user=user,
                                  host=host,
                                  port=port,
                                  auth=AuthMethod.GENERATE_KEY,
                                  key=KeyType.RSA,
                                  install_key=True,
                                  setup_actions=setup_actions,
                                  use_jupyter_lab=use_jupyter_lab)
            node = cluster.get_access_node()
            node.connect(password=password)
        else:
            cluster = add_cluster(name=cluster_name,
                                  user=user,
                                  host=host,
                                  port=port,
                                  auth=AuthMethod.PRIVATE_KEY,
                                  key=self.key_path,
                                  install_key=False,
                                  setup_actions=setup_actions,
                                  use_jupyter_lab=use_jupyter_lab)
            node = cluster.get_access_node()
            node.connect()

    def open_actions_file_dialog(self):
        """ Opens the window dialog that allows to select an actions file.
        """
        self.actions_file_name, _ = QFileDialog.getOpenFileName()
        if self.actions_file_name:
            self.ui.selected_file_path_browser.setText(self.actions_file_name)
            self.ui.selected_file_path_browser.setEnabled(True)
            self.ui.clear_actions_file_button.setEnabled(True)

    def clear_actions_file(self):
        """ Removes the selection of the actions file.
        """
        self.actions_file_name = None
        self.ui.selected_file_path_browser.setText("")
        self.ui.selected_file_path_browser.setEnabled(False)
        self.ui.clear_actions_file_button.setEnabled(False)

    def add_key(self):
        """ Opens the window dialog that allows to select a private key file.
        """
        self.key_path, _ = QFileDialog.getOpenFileName()
        if self.key_path:
            self.ui.selected_key_browser.setText(self.key_path)
            self.ui.selected_key_browser.setEnabled(True)
            self.ui.clear_key_button.setEnabled(True)

    def clear_key(self):
        """ Removes the selection of the private key file.
        """
        self.key_path = None
        self.ui.selected_key_browser.setText("")
        self.ui.selected_key_browser.setEnabled(False)
        self.ui.clear_key_button.setEnabled(False)



""" One of the helpers for the gui application.

    Similar modules: class:`.DataProvider`, :class:`.ParameterSaver`,
    :class:`.UiLoader`, :class:`.Worker`, :class:`.ConfigurationProvider`
"""
import json
import os


class NativeArgsSaver:
    """ Manages the native arguments.
    """

    def __init__(self):
        current_directory = os.path.dirname(os.path.abspath(__file__))
        self.filename = os.path.join(current_directory, '../../native_args.json')
        with open(self.filename) as json_file:
            self.native_args = json.load(json_file)

    def save(self, native_args):
        """ Saves the native arguments into a json file.

            :param native_args: Native arguments to be saved.
        """
        self.native_args = native_args
        with open(self.filename, 'w') as json_file:
            json.dump(native_args, json_file)

    def get_native_args(self):
        return self.native_args



""" One of the helper widgets.

    See: :class:`.ProgramInfoWindow`, :class:`.ShowLogsWindow`,
    :class:`.YesOrNoWindow`, :class:`.ShowJobsWindow`,
    :class:`.EditNativeArgumentsWindow`, :class:`.HelpWindow`
"""
from enum import Enum
from PyQt5.QtWidgets import QWidget, QMessageBox


class WindowType(Enum):
    """ Available popup window types.

        :attr:`.success`: Window that shows a success message.

        :attr:`.error`: Window that shows an error message.

    """
    success = 1
    error = 2


class PopUpWindow(QWidget):
    """ Helper widget popup window.
    """
    def __init__(self):
        super().__init__()
        self.box = QMessageBox(self)
        self.box.addButton(QMessageBox.Ok)
        self.box.setDefaultButton(QMessageBox.Ok)

    def show_message(self, message, window_type, error_info=None):
        """ Shows the popup window.

            :param message: Message to be shown.
            :param window_type: Type of the window.
            :param error_info: Additional information about the error.
        """
        if window_type == WindowType.success:
            self.box.setWindowTitle("Success")
            self.box.setIcon(QMessageBox.Information)
        elif window_type == WindowType.error:
            self.box.setWindowTitle("Error")
            self.box.setIcon(QMessageBox.Critical)
        else:
            self.box.setIcon(QMessageBox.NoIcon)

        self.box.setText(message)
        self.box.setDetailedText("")

        if error_info:
            self.box.setDetailedText(str(error_info))

        ret = self.box.exec_()
        if ret == QMessageBox.Ok:
            return
